import os
from subprocess import Popen, PIPE
import logger

class PullRequest():

    def __init__(self,
                 branch,
                 clone_url,
                 base_dir,
                 name,
                 upstream):
        self.branch = branch
        self.clone_url = clone_url
        self.base_dir = base_dir
        self.name = name
        self.upstream_url = upstream
        self.staging_dir = os.path.join(self.base_dir, self.name)

    def local_sha(self):
        os.chdir(self.staging_dir)
        args = ["git", "log", "--pretty='%H'", "HEAD^1.."]
        cmd = Popen(args, stdin=PIPE, stdout=PIPE, stderr=PIPE)
        output, err = cmd.communicate()
        return output

    def staged(self):
        return os.path.isfile(self.staging_dir)

    def behind(self, treeish):
        os.chdir(self.staging_dir)
        args = ["git", "log", "--oneline", "..{0}".format(treeish)]
        cmd = Popen(args, stdin=PIPE, stdout=PIPE, stderr=PIPE)
        output, err = cmd.communicate()
        return len(output.split('\n'))

    # A threshold for how many commits a branch can be behind upstream.
    # this should be a variable instead?
    def behind_threshold(self):
        return 10

    def up2date(self, treeish):
        return self.behind(treeish) <= self.behind_threshold()

    def rebase(self): #???
        logger.log("rebase origin/{0}".format(self.branch))
        os.chdir(self.staging_dir)
        args = ["git", "rebase", "origin/{0}".format(self.branch), "&>", "/dev/null"]
        cmd = Popen(args, stdin=PIPE, stdout=PIPE, stderr=PIPE)
        output, err = cmd.communicate()
        status = cmd.returncode
        self.update_submodules()
        logger.log("{0} is at revision {1}".format(self.branch, self.local_sha()))
        return status

    def reset_hard(self):
        os.chdir(self.staging_dir)
        args = ["git", "reset", "--hard", "origin/{0}".format(self.branch), "&>", "/dev/null"]
        cmd = Popen(args, stdin=PIPE, stdout=PIPE, stderr=PIPE)
        output, err = cmd.communicate()
        self.update_submodules()

    def fetch(self):
        logger.log("git fetch")
        os.chdir(self.staging_dir)
        self.add_upstream_remote()
        args = ["git", "fetch", "--all", "--prune", "&>", "/dev/null"]
        cmd = Popen(args, stdin=PIPE, stdout=PIPE, stderr=PIPE)
        output, err = cmd.communicate()
        pass

    def update_submodules(self):
        logger.log("update submodules in {0}".format(self.staging_dir))
        args = ["git", "submodule", "sync", "&>", "/dev/null"]
        cmd = Popen(args, stdin=PIPE, stdout=PIPE, stderr=PIPE)
        output, err = cmd.communicate()
        args = ["git", "submodule", "update", "--init", "&>", "/dev/null"]
        cmd = Popen(args, stdin=PIPE, stdout=PIPE, stderr=PIPE)
        output, err = cmd.communicate()
        pass
        
    def remote(self, s): # what is s ???
        os.chdir(self.staging_dir)
        args = ["git", "remote", "-v"]
        cmd = Popen(args, stdin=PIPE, stdout=PIPE, stderr=PIPE)
        output, err = cmd.communicate()
        remotes = output.split('\n')
        if remotes[-1] == '':
            remotes = remotes[:-1]
        urls = []
        for remote in remotes:
            remote = remote.split("\t")[-1].split(' ')[0]
            if remote not in urls:
                urls.append(remote)
        return s in urls

    def add_upstream_remote(self):
        os.chdir(self.staging_dir)
        logger.log("add upstream remote")
        args = ["git", "remote", "add", "upstream", self.upstream_url, "&>", "/dev/null"]
        cmd = Popen(args, stdin=PIPE, stdout=PIPE, stderr=PIPE)
        output, err = cmd.communicate()
        args = ["git", "fetch", "--prune", "upstream", "&>" "/dev/null"]
        cmd = Popen(args, stdin=PIPE, stdout=PIPE, stderr=PIPE)
        output, err = cmd.communicate()
        pass

    def clone(self):
        logger.log("clone to {0}".format(self.staging_dir))
        if not os.path.exists(self.base_dir):
            os.path.makedirs(self.base_dir)
        args = ["git", "clone", "-b", self.base_dir, self.clone_url, self.staging_dir, "&>", "/dev/null"]
        cmd = Popen(args, stdin=PIPE, stdout=PIPE, stderr=PIPE)
        output, err = cmd.communicate()
        os.chdir(self.staging_dir)
        self.add_upstream_remote()
        self.update_submodules()
        pass




import errno
import os
import signal
from git_timeout import GitTimeout

class timeout:

    def __init__(self, seconds, error_message=os.strerror(errno.ETIME)):
        self.seconds = seconds
        self.error_message = error_message

    def handle_timeout(self, signum, frame):
        raise GitTimeout(self.error_message)

    def __enter__(self):
        signal.signal(signal.SIGALRM, self.handle_timeout)
        signal.alarm(self.seconds)

    def __exit__(self, type, value, traceback):
        signal.alarm(0)




import os
import re
import shutil
import socket
import github3
import logger
import pull_request

class Autostager():

    def __init__(self):
        slug = self.repo_slug().split('/')
        self.owner = slug[0]
        self.repo = slug[1]
        self.access_token = os.environ.get('access_token')

    def access_token(self):
        return os.environ['access_token']

    def alphafy(self, a_string):
        pattern = "[^a-z0-9_]"
        return re.sub(pattern, "_", a_string, flags=re.IGNORECASE)


    def stage_upstream(self):
        default_branch = client.repository(self.owner, self.repo).default_branch
        logger.log("===> begin {0}".format(default_branch))
        p = pull_request.PullRequest(
            default_branch,
            self.authenticated_url("https://github.com/{0}".format(self.repo_slug())),
            self.base_dir(),
            default_branch,
            self.authenticated_url("https://github.com/{0}".format(self.repo_slug()))
        )
        if not p.staged():
            p.clone()
        p.fetch()
        if p.rebase():
            return
        self.client.repository(self.owner, self.repo).create_issue(
            "Failed to fast-forward {0} branch".format(default_branch),
            ":bangbang: This probably means somebody force-pushed to the branch."
        )
        pass



    def process_pull(self, pr):
        logger.log("===> {0} {1}".format(pr.number, self.staging_dir(pr)))
        p = pull_request.PullRequest(
            pr.head.ref,
            self.authenticated_url(pr.as_dict()['head']['repo']['clone_url']),
            self.base_dir(),
            self.clone_dir(pr),
            self.authenticated_url(pr.as_dict()['base']['repo']['clone_url']))
        if p.staged():
            p.fetch()
            if pr.head.sha != p.local_sha():
                p.reset_hard()
                add_comment = true
            else:
                logger.log("nothing to do on {0} {1}".format(pr.number, self.staging_dir(pr)))
                add_comment = false
            self.comment_or_close(p, pr, add_comment)
        else:
            p.clone()
            self.comment_or_close(p, pr)

    def comment_or_close(self, p, pr, add_comment = True):
        default_branch = pr.as_dict()['base']['repo']['default_branch']
        if p.up2date("upstream/{0}".format(default_branch)):
            if add_comment:
                comment = ":bell: Staged {0} at revision {1} on {2}"
                comment = comment.format(self.clone_dir(pr), p.local_sha(), socket.gethostname())
                pr.create_comment(comment)
                logger.log(comment)
            else:
                comment = ":boom: Unstaged since {0} is dangerously behind upstream"
                comment = comment.format(self.clone_dir(pr))
                shutil.rmtree(self.staging_dir(pr))
                pr.create_comment(comment)
                pr.close()
                logger.log(comment) 


    def authenticated_url(self, url):
        return re.sub(r"(^https://)", "https://" + self.access_token() + "@", url)


    def base_dir(self):
        return os.environ.get('base_dir') or '/opt/puppet/environments'


    def clone_dir(self, pr):
        alphafy(pr.head.label)


    def staging_dir(self, pr):
        os.path.join(self.base_dir(), self.clone_dir(pr))


    def repo_slug(self):
        return os.environ.get('repo_slug') # handle None return value if key does not exist?


    def client(self):
        return github3.login(token=access_token())


    def timeout_seconds(self):
        result = 120
        if os.environ.get('timeout'):
            result = int(os.environ['timeout'])
            assert result > 0, "timeout must be greater than zero seconds"
        return result


    def safe_dirs(self):
        return ['.', '..', 'production']

    def run(self):
        self.client()
        self.stage_upstream()
        prs = client.repository(self.owner, self.repo).pull_requests()
        new_clones = [self.clone_dir(pr) for pr in prs]
        if os.path.exists(self.base_dir()):
            discard_dirs = set(os.listdir(self.base_dir())) - set(self.safe_dirs()) - set(new_clones)
            discard_dirs = list(discard_dirs)
            discard_dirs = [os.path.join(self.base_dir(), d) for d in discard_dirs]
            for discard_dir in discard_dirs:
                logger.log("===> Unstage {0} since PR is closed.".format(discard_dir))
                shutil.rmtree(discard_dir) 
        with timeout(self.timeout_seconds()):
            for pr in prs:
                self.process_pull(pr)
        pass



from distutils.core import setup
setup(
  name = 'autostager',
  packages = ['autostager'],
  version = '0.1.2',
  description = 'Stage a directory based on Github pull request (e.g., dynamic puppet environments)',
  author = 'Jordan Facibene',
  author_email = 'jordan.facibene13@stjohns.edu',
  url = 'https://github.com/jfach/autostager',
  download_url = 'https://github.com/jfach/autostager/tarball/0.1.2',
  keywords = ['github', 'automation', 'staging'],
  classifiers = [],
  install_requires = ['github3']
)




from __future__ import print_function
import os
import sys
import syslog
import urllib
import datetime

def log(msg, level = syslog.LOG_INFO):
    msg = safe(msg)
    if os.environ.get('debug'):
        debugmsg = '{0} {1}'.format(datetime.datetime.now(), msg)
        print(debugmsg, file=sys.stderr)
    syslog.openlog('Autostager', syslog.LOG_PID, syslog.LOG_LOCAL0)
    syslog.syslog(level, msg)


def safe(str):
    return urllib.unquote_plus(str)
